# pyenchant
#
# Copyright (C) 2004-2008, Ryan Kelly
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# In addition, as a special exception, you are
# given permission to link the code of this program with
# non-LGPL Spelling Provider libraries (eg: a MSFT Office
# spell checker backend) and distribute linked combinations including
# the two.  You must obey the GNU Lesser General Public License in all
# respects for all of the code used other than said providers.  If you modify
# this file, you may extend this exception to your version of the
# file, but you are not obligated to do so.  If you do not wish to
# do so, delete this exception statement from your version.
#
# Major code cleanup and re-write thanks to Phil Mayes, 2007
#
"""

enchant.checker.wxSpellCheckerDialog: wxPython spellchecker interface

This module provides the class :py:class:`wxSpellCheckerDialog`, which provides
a wxPython dialog that can be used as an interface to a spell checking
session.  Currently it is intended as a proof-of-concept and demonstration
class, but it should be suitable for general-purpose use in a program.

The class must be given an :py:class:`enchant.checker.SpellChecker` object with
which to operate.  It can (in theory...) be used in modal and non-modal
modes.  Use `Show()` when operating on an array of characters as it will
modify the array in place, meaning other work can be done at the same
time.  Use `ShowModal()` when operating on a static string.

"""
_DOC_ERRORS = ["ShowModal"]

from typing import Any, Optional

import wx

from enchant.checker import SpellChecker


class wxSpellCheckerDialog(wx.Dialog):
    """Simple spellcheck dialog for wxPython

    This class implements a simple spellcheck interface for wxPython,
    in the form of a dialog.  It's intended mainly of an example of
    how to do this, although it should be useful for applications that
    just need a simple graphical spellchecker.

    To use, a :py:class:`SpellChecker` instance must be created and passed to the
    dialog before it is shown:

        >>> chkr = SpellChecker("en_AU",text)
        >>> dlg = wxSpellCheckerDialog(chkr,None,-1,"")
        >>> dlg.Show()

    This is most useful when the text to be checked is in the form of
    a character array, as it will be modified in place as the user
    interacts with the dialog.  For checking strings, the final result
    will need to be obtained from the `SpellChecker` object:

        >>> chkr = SpellChecker("en_AU",text)
        >>> dlg = wxSpellCheckerDialog(chkr,None,-1,"")
        >>> dlg.ShowModal()
        >>> text = chkr.get_text()

    Currently the checker must deal with strings of the same type as
    returned by wxPython - unicode or normal string depending on the
    underlying system.  This needs to be fixed, somehow...
    """

    _DOC_ERRORS = [
        "dlg",
        "chkr",
        "dlg",
        "chkr",
        "dlg",
        "dlg",
        "chkr",
        "dlg",
        "chkr",
        "dlg",
        "ShowModal",
        "dlg",
    ]

    # Remember dialog size across invocations by storing it on the class
    sz = (300, 70)

    def __init__(
        self,
        checker: SpellChecker,
        parent: Optional[Any] = None,
        id: int = -1,
        title: str = "Checking Spelling...",
        size=wxSpellCheckerDialog.sz,
        style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER,
    ) -> None:
        self._numContext = 40
        self._checker = checker
        self._buttonsEnabled = True
        self.error_text = wx.TextCtrl(
            self, -1, "", style=wx.TE_MULTILINE | wx.TE_READONLY | wx.TE_RICH
        )
        self.replace_text = wx.TextCtrl(self, -1, "", style=wx.TE_PROCESS_ENTER)
        self.replace_list = wx.ListBox(self, -1, style=wx.LB_SINGLE)
        self.InitLayout()
        wx.EVT_LISTBOX(self, self.replace_list.GetId(), self.OnReplSelect)
        wx.EVT_LISTBOX_DCLICK(self, self.replace_list.GetId(), self.OnReplace)

    def InitLayout(self) -> None:
        """Lay out controls and add buttons."""
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        txtSizer = wx.BoxSizer(wx.VERTICAL)
        btnSizer = wx.BoxSizer(wx.VERTICAL)
        replaceSizer = wx.BoxSizer(wx.HORIZONTAL)
        txtSizer.Add(
            wx.StaticText(self, -1, "Unrecognised Word:"), 0, wx.LEFT | wx.TOP, 5
        )
        txtSizer.Add(self.error_text, 1, wx.ALL | wx.EXPAND, 5)
        replaceSizer.Add(
            wx.StaticText(self, -1, "Replace with:"),
            0,
            wx.ALL | wx.ALIGN_CENTER_VERTICAL,
            5,
        )
        replaceSizer.Add(self.replace_text, 1, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        txtSizer.Add(replaceSizer, 0, wx.EXPAND, 0)
        txtSizer.Add(self.replace_list, 2, wx.ALL | wx.EXPAND, 5)
        sizer.Add(txtSizer, 1, wx.EXPAND, 0)
        self.buttons = []
        for label, action, tip in (
            ("Ignore", self.OnIgnore, "Ignore this word and continue"),
            (
                "Ignore All",
                self.OnIgnoreAll,
                "Ignore all instances of this word and continue",
            ),
            ("Replace", self.OnReplace, "Replace this word"),
            ("Replace All", self.OnReplaceAll, "Replace all instances of this word"),
            ("Add", self.OnAdd, "Add this word to the dictionary"),
            ("Done", self.OnDone, "Finish spell-checking and accept changes"),
        ):
            btn = wx.Button(self, -1, label)
            btn.SetToolTip(wx.ToolTip(tip))
            btnSizer.Add(btn, 0, wx.ALIGN_RIGHT | wx.ALL, 4)
            btn.Bind(wx.EVT_BUTTON, action)
            self.buttons.append(btn)
        sizer.Add(btnSizer, 0, wx.ALL | wx.EXPAND, 5)
        self.SetAutoLayout(True)
        self.SetSizer(sizer)
        sizer.Fit(self)

    def Advance(self) -> bool:
        """Advance to the next error.

        This method advances the SpellChecker to the next error, if
        any.  It then displays the error and some surrounding context,
        and well as listing the suggested replacements.
        """
        # Advance to next error, disable if not available
        try:
            self._checker.next()
        except StopIteration:
            self.EnableButtons(False)
            self.error_text.SetValue("")
            self.replace_list.Clear()
            self.replace_text.SetValue("")
            if self.IsModal():  # test needed for SetSpellChecker call
                # auto-exit when checking complete
                self.EndModal(wx.ID_OK)
            return False
        self.EnableButtons()
        # Display error context with erroneous word in red.
        # Restoring default style was misbehaving under win32, so
        # I am forcing the rest of the text to be black.
        self.error_text.SetValue("")
        self.error_text.SetDefaultStyle(wx.TextAttr(wx.BLACK))
        lContext = self._checker.leading_context(self._numContext)
        self.error_text.AppendText(lContext)
        self.error_text.SetDefaultStyle(wx.TextAttr(wx.RED))
        self.error_text.AppendText(self._checker.word)
        self.error_text.SetDefaultStyle(wx.TextAttr(wx.BLACK))
        tContext = self._checker.trailing_context(self._numContext)
        self.error_text.AppendText(tContext)
        # Display suggestions in the replacements list
        suggs = self._checker.suggest()
        self.replace_list.Set(suggs)
        self.replace_text.SetValue(suggs and suggs[0] or "")
        return True

    def EnableButtons(self, state: bool = True) -> None:
        """Enable the checking-related buttons"""
        if state != self._buttonsEnabled:
            for btn in self.buttons[:-1]:
                btn.Enable(state)
            self._buttonsEnabled = state

    def GetRepl(self) -> str:
        """Get the chosen replacement string."""
        repl = self.replace_text.GetValue()
        return repl

    def OnAdd(self, evt: Any) -> None:
        """Callback for the "add" button."""
        self._checker.add()
        self.Advance()

    def OnDone(self, evt: Any) -> None:
        """Callback for the "close" button."""
        wxSpellCheckerDialog.sz = self.error_text.GetSizeTuple()
        if self.IsModal():
            self.EndModal(wx.ID_OK)
        else:
            self.Close()

    def OnIgnore(self, evt: Any) -> None:
        """Callback for the "ignore" button.
        This simply advances to the next error.
        """
        self.Advance()

    def OnIgnoreAll(self, evt: Any) -> None:
        """Callback for the "ignore all" button."""
        self._checker.ignore_always()
        self.Advance()

    def OnReplace(self, evt: Any) -> None:
        """Callback for the "replace" button."""
        repl = self.GetRepl()
        if repl:
            self._checker.replace(repl)
        self.Advance()

    def OnReplaceAll(self, evt: Any) -> None:
        """Callback for the "replace all" button."""
        repl = self.GetRepl()
        self._checker.replace_always(repl)
        self.Advance()

    def OnReplSelect(self, evt: Any) -> None:
        """Callback when a new replacement option is selected."""
        sel = self.replace_list.GetSelection()
        if sel == -1:
            return
        opt = self.replace_list.GetString(sel)
        self.replace_text.SetValue(opt)


def _test():
    class TestDialog(wxSpellCheckerDialog):
        def __init__(self, *args):
            super().__init__(*args)
            wx.EVT_CLOSE(self, self.OnClose)

        def OnClose(self, evnt):
            print(["AFTER:", dlg._checker.get_text()])
            self.Destroy()

    from enchant.checker import SpellChecker

    text = "This is sme text with a fw speling errors in it. Here are a fw more to tst it ut."
    print(["BEFORE:", text])
    chkr = SpellChecker("en_US", text)

    app = wx.App(False)
    dlg = TestDialog(chkr)
    dlg.Show()
    app.MainLoop()


if __name__ == "__main__":
    _test()



# pyenchant
#
# Copyright (C) 2004-2011 Ryan Kelly
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# In addition, as a special exception, you are
# given permission to link the code of this program with
# non-LGPL Spelling Provider libraries (eg: a MSFT Office
# spell checker backend) and distribute linked combinations including
# the two.  You must obey the GNU Lesser General Public License in all
# respects for all of the code used other than said providers.  If you modify
# this file, you may extend this exception to your version of the
# file, but you are not obligated to do so.  If you do not wish to
# do so, delete this exception statement from your version.
#
"""

pypwl:  pure-python personal word list in the style of Enchant
==============================================================

This module provides a pure-python version of the personal word list
functionality found in the spellchecking package Enchant.  While the
same effect can be achieved (with better performance) using the python
bindings for Enchant, it requires a C extension.

This pure-python implementation uses the same algorithm but without any
external dependencies or C code (in fact, it was the author's original
prototype for the C version found in Enchant).

"""


import os
import warnings
from typing import Dict, Iterable, Iterator, List, Optional  # noqa F401


class Trie:
    """Class implementing a trie-based dictionary of words.

    A Trie is a recursive data structure storing words by their prefix.
    "Fuzzy matching" can be done by allowing a certain number of missteps
    when traversing the Trie.
    """

    def __init__(self, words: Iterable[str] = ()) -> None:
        self._eos = False  # whether I am the end of a word
        self._keys = {}  # type: Dict[str, Trie] # letters at this level of the trie
        for w in words:
            self.insert(w)

    def insert(self, word: str) -> None:
        if word == "":
            self._eos = True
        else:
            key = word[0]
            try:
                subtrie = self[key]
            except KeyError:
                subtrie = Trie()
                self[key] = subtrie
            subtrie.insert(word[1:])

    def remove(self, word: str) -> None:
        if word == "":
            self._eos = False
        else:
            key = word[0]
            try:
                subtrie = self[key]
            except KeyError:
                pass
            else:
                subtrie.remove(word[1:])

    def search(self, word: str, nerrs: int = 0) -> List[str]:
        """Search for the given word, possibly making errors.

        This method searches the trie for the given `word`, making
        precisely `nerrs` errors.  It returns a list of words found.
        """
        res = []  # type: List[str]
        # Terminate if we've run out of errors
        if nerrs < 0:
            return res
        # Precise match at the end of the word
        if nerrs == 0 and word == "":
            if self._eos:
                res.append("")
        # Precisely match word[0]
        try:
            subtrie = self[word[0]]
            subres = subtrie.search(word[1:], nerrs)
            for w in subres:
                w2 = word[0] + w
                if w2 not in res:
                    res.append(w2)
        except (IndexError, KeyError):
            pass
        # match with deletion of word[0]
        try:
            subres = self.search(word[1:], nerrs - 1)
            for w in subres:
                if w not in res:
                    res.append(w)
        except (IndexError,):
            pass
        # match with insertion before word[0]
        try:
            for k in self._keys:
                subres = self[k].search(word, nerrs - 1)
                for w in subres:
                    w2 = k + w
                    if w2 not in res:
                        res.append(w2)
        except (IndexError, KeyError):
            pass
        # match on substitution of word[0]
        try:
            for k in self._keys:
                subres = self[k].search(word[1:], nerrs - 1)
                for w in subres:
                    w2 = k + w
                    if w2 not in res:
                        res.append(w2)
        except (IndexError, KeyError):
            pass
        # All done!
        return res

    search._DOC_ERRORS = ["nerrs"]  # type: ignore

    def __getitem__(self, key: str) -> "Trie":
        return self._keys[key]

    def __setitem__(self, key: str, val: "Trie") -> None:
        self._keys[key] = val

    def __iter__(self) -> Iterator[str]:
        if self._eos:
            yield ""
        for k in self._keys:
            for w2 in self._keys[k]:
                yield k + w2


class PyPWL:
    """Pure-python implementation of Personal Word List dictionary.
    This class emulates the PWL objects provided by PyEnchant, but
    implemented purely in python.
    """

    def __init__(self, pwl: Optional[str] = None) -> None:
        """PyPWL constructor.
        This method takes as its only argument the name of a file
        containing the personal word list, one word per line.  Entries
        will be read from this file, and new entries will be written to
        it automatically.

        If `pwl` is not specified or None, the list is maintained in
        memory only.
        """
        self.provider = None
        self._words = Trie()
        if pwl is not None:
            self.pwl = os.path.abspath(pwl)  # type: Optional[str]
            self.tag = self.pwl
            pwl_f = open(pwl)
            for ln in pwl_f:
                word = ln.strip()
                self.add_to_session(word)
            pwl_f.close()
        else:
            self.pwl = None
            self.tag = "PyPWL"

    def check(self, word: str) -> bool:
        """Check spelling of a word.

        This method takes a word in the dictionary language and returns
        `True` if it is correctly spelled, and `False` otherwise.
        """
        res = self._words.search(word)
        return bool(res)

    def suggest(self, word: str) -> List[str]:
        """Suggest possible spellings for a word.

        This method tries to guess the correct spelling for a given
        word, returning the possibilities in a list.
        """
        limit = 10
        maxdepth = 5
        # Iterative deepening until we get enough matches
        depth = 0
        res = self._words.search(word, depth)
        while len(res) < limit and depth < maxdepth:
            depth += 1
            for w in self._words.search(word, depth):
                if w not in res:
                    res.append(w)
        # Limit number of suggs
        return res[:limit]

    def add(self, word: str) -> None:
        """Add a word to the user's personal dictionary.
        For a PWL, this means appending it to the file.
        """
        if self.pwl is not None:
            pwl_f = open(self.pwl, "a")
            pwl_f.write("%s\n" % (word.strip(),))
            pwl_f.close()
        self.add_to_session(word)

    def add_to_pwl(self, word: str) -> None:
        """Add a word to the user's personal dictionary.
        For a PWL, this means appending it to the file.
        """
        warnings.warn(
            "PyPWL.add_to_pwl is deprecated, please use PyPWL.add",
            category=DeprecationWarning,
            stacklevel=2,
        )
        self.add(word)

    def remove(self, word: str) -> None:
        """Add a word to the user's personal exclude list."""
        # There's no exclude list for a stand-alone PWL.
        # Just remove it from the list.
        self._words.remove(word)
        if self.pwl is not None:
            pwl_f = open(self.pwl, "wt")
            for w in self._words:
                pwl_f.write("%s\n" % (w.strip(),))
            pwl_f.close()

    def add_to_session(self, word: str) -> None:
        """Add a word to the session list."""
        self._words.insert(word)

    def store_replacement(self, mis: str, cor: str) -> None:
        """Store a replacement spelling for a miss-spelled word.

        This method makes a suggestion to the spellchecking engine that the
        miss-spelled word `mis` is in fact correctly spelled as `cor`.  Such
        a suggestion will typically mean that `cor` appears early in the
        list of suggested spellings offered for later instances of `mis`.
        """
        # Too much work for this simple spellchecker
        pass

    store_replacement._DOC_ERRORS = ["mis", "mis"]  # type: ignore

    def is_added(self, word: str) -> bool:
        """Check whether a word is in the personal word list."""
        return self.check(word)

    def is_removed(self, word: str) -> bool:
        """Check whether a word is in the personal exclude list."""
        return False

    #  No-op methods to support internal use as a Dict() replacement

    def _check_this(self, msg: str) -> None:
        pass

    def _free(self) -> None:
        pass



# pyenchant
#
# Copyright (C) 2004-2008, Ryan Kelly
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# In addition, as a special exception, you are
# given permission to link the code of this program with
# non-LGPL Spelling Provider libraries (eg: a MSFT Office
# spell checker backend) and distribute linked combinations including
# the two.  You must obey the GNU Lesser General Public License in all
# respects for all of the code used other than said providers.  If you modify
# this file, you may extend this exception to your version of the
# file, but you are not obligated to do so.  If you do not wish to
# do so, delete this exception statement from your version.
#
"""
enchant.errors:  Error class definitions for the enchant library
================================================================

All error classes are defined in this separate sub-module, so that they
can safely be imported without causing circular dependencies.

"""


class Error(Exception):
    """Base exception class for the enchant module."""

    pass


class DictNotFoundError(Error):
    """Exception raised when a requested dictionary could not be found."""

    pass


class TokenizerNotFoundError(Error):
    """Exception raised when a requested tokenizer could not be found."""

    pass


class DefaultLanguageNotFoundError(Error):
    """Exception raised when a default language could not be found."""

    pass
